# -*- coding: utf-8 -*-

"""Top-level package for {{ cookiecutter.project_slug }}."""

__author__ = '{{ cookiecutter.full_name.replace("\'", "\\\'") }}'
__email__ = '{{ cookiecutter.email }}'

import {{ cookiecutter.package_slug }}.load_data
import {{ cookiecutter.package_slug }}.features
