from ballet import Feature


input = None  # TODO - str or list of str
transformer = None  # TODO - function, transformer-like, or list thereof
name = None  # TODO - str
feature = Feature(input, transformer, name)
